from yarl import URL

from ..base_functions.base_functions import log, logger, make_title_safe, get_db_path, get_filename_and_ext
from ..base_functions.data_classes import AlbumItem, MediaItem
from ..base_functions.error_classes import NoExtensionFailure
from ..base_functions.sql_helper import SQLHelper
from ..client.client import ScrapeSession


class XBunkrCrawler:
    def __init__(self, quiet: bool, SQL_Helper: SQLHelper):
        self.quiet = quiet
        self.SQL_Helper = SQL_Helper

    async def fetch(self, session: ScrapeSession, url: URL):
        """Director for XBunkr scraping"""
        album_obj = AlbumItem("Loose XBunkr Files", [])

        try:
            if "media" in url.host:
                url_path = await get_db_path(url)
                complete = await self.SQL_Helper.check_complete_singular("xbunkr", url_path)
                filename, ext = await get_filename_and_ext(url.name)
                media_item = MediaItem(url, url, complete, filename, ext, filename)
                await album_obj.add_media(media_item)

            else:
                soup = await session.get_BS4(url)
                links = soup.select("a[class=image]")
                title = await make_title_safe(soup.select_one("h1[id=title]").text)
                title = title.strip()
                await album_obj.set_new_title(title)
                for link in links:
                    link = URL(link.get('href'))
                    try:
                        filename, ext = await get_filename_and_ext(link.name)
                    except NoExtensionFailure:
                        logger.debug("Couldn't get extension for %s", str(link))
                        continue
                    url_path = await get_db_path(link)
                    complete = await self.SQL_Helper.check_complete_singular("xbunkr", url_path)
                    media_item = MediaItem(link, url, complete, filename, ext, filename)
                    await album_obj.add_media(media_item)

        except Exception as e:
            logger.debug("Error encountered while handling %s", str(url), exc_info=True)
            await log("Error scraping " + str(url), quiet=self.quiet)
            logger.debug(e)

        await self.SQL_Helper.insert_album("xbunkr", "", album_obj)
        await log(f"[green]Finished: {str(url)}[/green]", quiet=self.quiet)
        return album_obj



import atexit
import logging
import sqlite3
from pathlib import Path

from cyberdrop_dl.base_functions.base_functions import get_db_path
from cyberdrop_dl.base_functions.data_classes import AlbumItem, CascadeItem, DomainItem


class SQLHelper:
    """This class is responsible for handling SQL operations"""
    def __init__(self, ignore_history, ignore_cache, download_history):
        self.ignore_history = ignore_history
        self.ignore_cache = ignore_cache
        self.download_history = download_history
        self.conn = None
        self.curs = None

        self.old_history = False
        # Close the sql connection when the program exits
        atexit.register(self.exit_handler)

    async def sql_initialize(self):
        """Initializes the SQL connection, and makes sure necessary tables exist"""
        self.conn = sqlite3.connect(self.download_history)
        self.curs = self.conn.cursor()

        await self.check_old_history()

        await self.pre_allocate()
        await self.create_media_history()
        await self.create_coomeno_history()

    async def check_old_history(self):
        """Checks whether V3 history exists"""
        try:
            self.curs.execute("""SELECT name FROM sqlite_schema WHERE type='table' AND name='downloads'""")
            sql_file_check = self.curs.fetchone()
        except Exception:
            self.curs.execute("""SELECT name FROM sqlite_master WHERE type='table' AND name='downloads'""")
            sql_file_check = self.curs.fetchone()
        if sql_file_check:
            self.old_history = True

    async def create_media_history(self):
        """We create the download history tables here"""
        create_table_query = """CREATE TABLE IF NOT EXISTS media (
                                                    domain TEXT,
                                                    url_path TEXT,
                                                    album_path TEXT,
                                                    referer TEXT,
                                                    download_path TEXT,
                                                    download_filename TEXT,
                                                    original_filename TEXT,
                                                    completed INTEGER NOT NULL,
                                                    PRIMARY KEY (url_path, original_filename)
                                                );"""
        create_temp_download_name_query = """CREATE TABLE IF NOT EXISTS downloads_temp (
                                                                downloaded_filename TEXT
                                                            );"""
        temp_truncate_query = """DELETE FROM downloads_temp;"""

        self.curs.execute(create_table_query)
        self.conn.commit()
        self.curs.execute(create_temp_download_name_query)
        self.conn.commit()
        self.curs.execute(temp_truncate_query)
        self.conn.commit()

    async def create_coomeno_history(self):
        """Creates the cache table for coomeno"""
        create_table_query = """CREATE TABLE IF NOT EXISTS coomeno (
                                                            url_path TEXT,
                                                            post_data BLOB,
                                                            PRIMARY KEY (url_path)
                                                        );"""
        self.curs.execute(create_table_query)
        self.conn.commit()

    async def pre_allocate(self):
        """We pre-allocate 50MB of space to the SQL file just in case the user runs out of disk space"""
        pre_alloc = "CREATE TABLE IF NOT EXISTS t(x);"
        pre_alloc2 = "INSERT INTO t VALUES(zeroblob(50*1024*1024));"  # 50 mb
        drop_pre = "DROP TABLE t;"
        check_prealloc = "PRAGMA freelist_count;"

        self.curs.execute(check_prealloc)
        free = self.curs.fetchone()[0]
        if free <= 1024:
            self.curs.execute(pre_alloc)
            self.conn.commit()
            self.curs.execute(pre_alloc2)
            self.conn.commit()
            self.curs.execute(drop_pre)
            self.conn.commit()

    """Temp Table Operations"""

    async def get_temp_names(self):
        """Gets the list of temp filenames"""
        self.curs.execute("SELECT downloaded_filename FROM downloads_temp;")
        filenames = self.curs.fetchall()
        filenames = list(sum(filenames, ()))
        return filenames

    async def sql_insert_temp(self, downloaded_filename):
        """Inserts a temp filename into the downloads_temp table"""
        self.curs.execute("""INSERT OR IGNORE INTO downloads_temp VALUES (?)""", (downloaded_filename,))
        self.conn.commit()

    """Coomeno Table Operations"""

    async def insert_blob(self, blob: str, url_path: str):
        """Inserts the post content into coomeno"""
        self.curs.execute("""INSERT OR IGNORE INTO coomeno VALUES (?, ?)""", (url_path, blob,))
        self.conn.commit()

    async def get_blob(self, url_path: str):
        """returns the post content for a given coomeno post url"""
        if self.ignore_cache:
            return None
        self.curs.execute("""SELECT post_data FROM coomeno WHERE url_path = ?""", (url_path,))
        sql_file_check = self.curs.fetchone()
        if sql_file_check:
            return sql_file_check[0]
        return None

    """Download Table Operations"""

    async def insert_media(self, domain: str, url_path: str, album_path: str, referer: str, download_path: str,
                           download_filename: str, original_filename: str, completed: int):
        """Inserts a media entry into the media table"""
        self.curs.execute("""INSERT OR IGNORE INTO media VALUES (?, ?, ?, ?, ?, ?, ?, ?)""",
                          (domain, url_path, album_path, referer, download_path, download_filename, original_filename, completed,))
        self.conn.commit()

    async def insert_album(self, domain: str, album_path: str, album: AlbumItem):
        """Inserts an albums media into the media table"""
        if album.media:
            for media in album.media:
                url_path = await get_db_path(media.url, domain)
                self.curs.execute("""INSERT OR IGNORE INTO media VALUES (?, ?, ?, ?, ?, ?, ?, ?)""",
                                  (domain, url_path, album_path, str(media.referer), "", "", media.filename, 0,))
        self.conn.commit()

    async def insert_domain(self, domain_name: str, album_path: str, domain: DomainItem):
        """Inserts a domains media into the media table"""
        if domain.albums:
            for title, album in domain.albums.items():
                for media in album.media:
                    url_path = await get_db_path(media.url, domain_name)
                    self.curs.execute("""INSERT OR IGNORE INTO media VALUES (?, ?, ?, ?, ?, ?, ?, ?)""",
                                      (domain_name, url_path, album_path, str(media.referer), "", "",
                                       media.filename, 0,))
        self.conn.commit()

    async def insert_cascade(self, cascade: CascadeItem):
        """Inserts a cascades media into the media table"""
        if not await cascade.is_empty():
            for domain, domain_obj in cascade.domains.items():
                for title, album_obj in domain_obj.albums.items():
                    for media in album_obj.media:
                        url_path = await get_db_path(media.url, domain)
                        self.curs.execute("""INSERT OR IGNORE INTO media VALUES (?, ?, ?, ?, ?, ?, ?, ?)""",
                                          (domain, url_path, media.referer.path, str(media.referer), "",
                                           "", media.filename, 0,))
        self.conn.commit()

    async def get_downloaded_filename(self, url_path, filename):
        """Gets downloaded filename given the url path and original filename"""
        self.curs.execute("""SELECT download_filename FROM media WHERE url_path = ? and original_filename = ?""",
                          (url_path, filename,))
        sql_file_check = self.curs.fetchone()
        if sql_file_check:
            return sql_file_check[0]
        return None

    async def sql_check_old_existing(self, url_path):
        """Checks the V3 history table for completed if it exists"""
        if not self.old_history:
            return False
        if self.ignore_history:
            return False
        self.curs.execute("""SELECT completed FROM downloads WHERE path = ?""", (url_path,))
        sql_file_check = self.curs.fetchone()
        return sql_file_check and sql_file_check[0] == 1

    async def check_complete_singular(self, domain, url_path):
        """Checks whether an individual file has completed given its domain and url path"""
        if self.ignore_history:
            return False
        self.curs.execute("""SELECT completed FROM media WHERE domain = ? and url_path = ?""", (domain, url_path,))
        sql_file_check = self.curs.fetchone()
        if not sql_file_check:
            return False
        elif sql_file_check[0] == 0:
            return False
        else:
            return True

    """Downloader Operations"""

    async def check_filename(self, filename):
        """Checks whether an individual exists in the DB given its filename"""
        self.curs.execute("""SELECT EXISTS(SELECT 1 FROM media WHERE download_filename = ?)""", (filename, ))
        sql_check = self.curs.fetchone()[0]
        return sql_check == 1

    async def update_pre_download(self, path: Path, filename: str, url_path: str, original_filename: str):
        """Update the media entry pre-download"""
        self.curs.execute("""UPDATE media SET download_path = ?, download_filename = ? WHERE url_path = ? 
        AND original_filename = ?""", (str(path), filename, url_path, original_filename,))
        self.conn.commit()

    async def mark_complete(self, url_path: str, original_filename: str):
        """Update the media entry post-download"""
        self.curs.execute("""UPDATE media SET completed = 1 WHERE url_path = ? AND original_filename = ?""", (url_path, original_filename,))
        self.conn.commit()

    def exit_handler(self):
        """Exit handler on unexpected exits"""
        try:
            self.conn.commit()
            self.conn.close()
        except Exception as e:
            logging.debug(f"Failed to close sqlite database connection: {str(e)}")
        else:
            logging.debug("Successfully closed sqlite database connection")



from __future__ import annotations

import copy
from pathlib import Path

import yaml

from cyberdrop_dl.base_functions.base_functions import log
from cyberdrop_dl.base_functions.config_schema import config_default, authentication_args, files_args, \
    jdownloader_args, runtime_args, forum_args, ignore_args, ratelimiting_args, sorting_args, progress_args


def create_config(config: Path, passed_args=None, remake=None, enabled=False):
    """Creates the default config file, or remakes it with passed arguments"""
    if config.is_file() and not remake:
        validate_config(config)
        return

    config_data = config_default
    if passed_args:
        config_data[0]["Configuration"]["Apply_Config"] = enabled
        for arg in authentication_args:
            if arg in passed_args.keys():
                config_data[0]["Configuration"]["Authentication"][arg] = passed_args[arg]
        for arg in files_args:
            if arg in passed_args.keys():
                config_data[0]["Configuration"]["Files"][arg] = str(passed_args[arg])
        for arg in ignore_args:
            if arg in passed_args.keys():
                config_data[0]["Configuration"]["Ignore"][arg] = str(passed_args[arg])
        for arg in forum_args:
            if arg in passed_args.keys():
                config_data[0]["Configuration"]["Forum_Options"][arg] = passed_args[arg]
        for arg in jdownloader_args:
            if arg in passed_args.keys():
                config_data[0]["Configuration"]["JDownloader"][arg] = passed_args[arg]
        for arg in progress_args:
            if arg in passed_args.keys():
                config_data[0]["Configuration"]["Progress_Options"][arg] = passed_args[arg]
        for arg in ratelimiting_args:
            if arg in passed_args.keys():
                config_data[0]["Configuration"]["Ratelimiting"][arg] = passed_args[arg]
        for arg in runtime_args:
            if arg in passed_args.keys():
                config_data[0]["Configuration"]["Runtime"][arg] = passed_args[arg]
        for arg in sorting_args:
            if arg in passed_args.keys():
                if arg == "sort_directory":
                    config_data[0]["Configuration"]["Sorting"][arg] = str(passed_args[arg])
                else:
                    config_data[0]["Configuration"]["Sorting"][arg] = passed_args[arg]

    with open(config, 'w') as yamlfile:
        yaml.dump(config_data, yamlfile)
    return


def validate_config(config: Path):
    """Validates the existing config file"""
    with open(config, "r") as yamlfile:
        data = yaml.load(yamlfile, Loader=yaml.FullLoader)
    try:
        data = data[0]["Configuration"]
        enabled = data["Apply_Config"]
        recreate = 0

        if not set(authentication_args).issubset(set(data['Authentication'].keys())):
            recreate = 1
        if not set(files_args).issubset(set(data['Files'].keys())):
            recreate = 1
        if not set(forum_args).issubset(set(data['Forum_Options'].keys())):
            recreate = 1
        if not set(ignore_args).issubset(set(data['Ignore'].keys())):
            recreate = 1
        if not set(jdownloader_args).issubset(set(data['JDownloader'].keys())):
            recreate = 1
        if not set(progress_args).issubset(set(data['Progress_Options'].keys())):
            recreate = 1
        if not set(ratelimiting_args).issubset(set(data['Ratelimiting'].keys())):
            recreate = 1
        if not set(runtime_args).issubset(set(data['Runtime'].keys())):
            recreate = 1
        if not set(sorting_args).issubset(set(data['Sorting'].keys())):
            recreate = 1

        if recreate:
            config.unlink()

            args = {}
            args_list = [data['Authentication'], data['Files'], data['Forum_Options'], data['Ignore'],
                         data['JDownloader'], data['Progress_Options'], data['Ratelimiting'], data['Runtime'],
                         data['Sorting']]
            for dic in args_list:
                args.update(dic)
            create_config(config, args, True, enabled)

    except (KeyError, TypeError):
        config.unlink()
        create_config(config, None, True)


def run_args(config: Path, cmd_arg: dict):
    """Returns the proper runtime arguments based on the config and command line arguments"""
    create_config(config, cmd_arg)

    with open(config, "r") as yamlfile:
        data = yaml.load(yamlfile, Loader=yaml.FullLoader)
    data = data[0]["Configuration"]
    if data['Apply_Config']:
        for file, path in data['Files'].items():
            data['Files'][file] = Path(path)
        data['Sorting']['sort_directory'] = Path(data['Sorting']['sort_directory'])
        return data

    config_data = config_default[0]["Configuration"]
    for arg in authentication_args:
        if arg in cmd_arg.keys():
            config_data["Authentication"][arg] = cmd_arg[arg]
    for arg in files_args:
        if arg in cmd_arg.keys():
            config_data["Files"][arg] = cmd_arg[arg]
    for arg in forum_args:
        if arg in cmd_arg.keys():
            config_data["Forum_Options"][arg] = cmd_arg[arg]
    for arg in ignore_args:
        if arg in cmd_arg.keys():
            config_data["Ignore"][arg] = cmd_arg[arg]
    for arg in jdownloader_args:
        if arg in cmd_arg.keys():
            config_data["JDownloader"][arg] = cmd_arg[arg]
    for arg in progress_args:
        if arg in cmd_arg.keys():
            config_data["Progress_Options"][arg] = cmd_arg[arg]
    for arg in ratelimiting_args:
        if arg in cmd_arg.keys():
            config_data["Ratelimiting"][arg] = cmd_arg[arg]
    for arg in runtime_args:
        if arg in cmd_arg.keys():
            config_data["Runtime"][arg] = cmd_arg[arg]
    for arg in sorting_args:
        if arg in cmd_arg.keys():
            config_data["Sorting"][arg] = cmd_arg[arg]
    return config_data


async def document_args(args: dict):
    """We document the runtime arguments for debugging and troubleshooting, redacting sensitive information"""
    print_args = copy.deepcopy(args)
    print_args['Authentication']['xbunker_password'] = '!REDACTED!' if args['Authentication']['xbunker_password'] is not None else None
    print_args['Authentication']['socialmediagirls_password'] = '!REDACTED!' if args['Authentication']['socialmediagirls_password'] is not None else None
    print_args['Authentication']['simpcity_password'] = '!REDACTED!' if args['Authentication']['simpcity_password'] is not None else None
    print_args['Authentication']['pixeldrain_api_key'] = '!REDACTED!' if args['Authentication']['pixeldrain_api_key'] is not None else None
    print_args['JDownloader']['jdownloader_password'] = '!REDACTED!' if args['JDownloader']['jdownloader_password'] is not None else None

    await log("Starting Cyberdrop-DL")
    await log(f"Using authentication arguments: {print_args['Authentication']}", quiet=True)
    await log(f"Using file arguments: {print_args['Files']}", quiet=True)
    await log(f"Using forum option arguments: {print_args['Forum_Options']}", quiet=True)
    await log(f"Using ignore arguments: {print_args['Ignore']}", quiet=True)
    await log(f"Using jdownloader arguments: {print_args['JDownloader']}", quiet=True)
    await log(f"Using progress option arguments: {print_args['Progress_Options']}", quiet=True)
    await log(f"Using ratelimiting arguments: {print_args['Ratelimiting']}", quiet=True)
    await log(f"Using runtime arguments: {print_args['Runtime']}", quiet=True)
    await log(f"Using sorting arguments: {print_args['Sorting']}", quiet=True)



from yarl import URL

from ..base_functions.base_functions import log, logger, get_filename_and_ext, get_db_path
from ..base_functions.data_classes import AlbumItem, MediaItem
from ..base_functions.error_classes import NoExtensionFailure
from ..base_functions.sql_helper import SQLHelper
from ..client.client import ScrapeSession


class PixelDrainCrawler:
    def __init__(self, quiet: bool, SQL_Helper: SQLHelper):
        self.quiet = quiet
        self.SQL_Helper = SQL_Helper
        self.api = URL('https://pixeldrain.com/api/')

    async def fetch(self, session: ScrapeSession, url: URL):
        """Director for pixeldrain scraping"""
        await log(f"[green]Starting: {str(url)}[/green]", quiet=self.quiet)
        album_obj = AlbumItem("Loose Pixeldrain Files", [])

        identifier = str(url).split('/')[-1]
        if url.parts[1] == 'l':
            await album_obj.set_new_title(url.name)
            media_items = await self.get_listings(session, identifier, url)
            if media_items:
                for media_item in media_items:
                    await album_obj.add_media(media_item)
        else:
            link = await self.create_download_link(identifier)
            url_path = await get_db_path(link)
            complete = await self.SQL_Helper.check_complete_singular("anonfiles", url_path)
            filename, ext = await get_filename_and_ext(await self.get_file_name(session, identifier))
            media_item = MediaItem(link, url, complete, filename, ext, filename)
            await album_obj.add_media(media_item)

        await self.SQL_Helper.insert_album("pixeldrain", url.path, album_obj)
        await log(f"[green]Finished: {str(url)}[/green]", quiet=self.quiet)
        return album_obj

    async def get_listings(self, session: ScrapeSession, identifier: str, url: URL):
        """Handles album scraping"""
        media_items = []
        try:
            content = await session.get_json((self.api / "list" / identifier))
            for file in content['files']:
                link = await self.create_download_link(file['id'])
                try:
                    filename, ext = await get_filename_and_ext(file['name'])
                except NoExtensionFailure:
                    logger.debug("Couldn't get extension for %s", str(link))
                    continue
                url_path = await get_db_path(link)
                complete = await self.SQL_Helper.check_complete_singular("pixeldrain", url_path)
                media_item = MediaItem(link, url, complete, filename, ext, filename)
                media_items.append(media_item)
            return media_items

        except Exception as e:
            logger.debug("Error encountered while handling %s", str(url), exc_info=True)
            await log(f"[red]Error: {str(url)}[/red]", quiet=self.quiet)
            logger.debug(e)

    async def get_file_name(self, session: ScrapeSession, identifier: str):
        """Gets filename for the given file identifier"""
        content = await session.get_json((self.api / 'file' / identifier / 'info'))
        filename = content['name']
        return filename

    async def create_download_link(self, file: str):
        """Gets download links for the file given"""
        final_url = (self.api / 'file' / file).with_query('download')
        return final_url



import shutil
from base64 import b64encode
from http import HTTPStatus
from pathlib import Path

from cyberdrop_dl.base_functions.base_functions import FILE_FORMATS
from cyberdrop_dl.base_functions.data_classes import MediaItem


async def allowed_filetype(media: MediaItem, block_images: bool, block_video: bool, block_audio: bool, block_other: bool):
    """Checks whether the enclosed file is allowed to be downloaded"""
    ext = media.ext
    if block_images:
        if ext in FILE_FORMATS["Images"]:
            return False
    if block_video:
        if ext in FILE_FORMATS["Videos"]:
            return False
    if block_audio:
        if ext in FILE_FORMATS["Audio"]:
            return False
    if block_other:
        if ext not in FILE_FORMATS["Images"] and ext not in FILE_FORMATS["Videos"] and ext not in FILE_FORMATS["Audio"]:
            return False
    return True


async def basic_auth(username, password):
    token = b64encode(f"{username}:{password}".encode('utf-8')).decode("ascii")
    return f'Basic {token}'


async def check_free_space(required_space_gb: int, download_directory: Path) -> bool:
    """Checks if there is enough free space on the drive to continue operating"""
    free_space = shutil.disk_usage(download_directory.parent).free
    free_space_gb = free_space / 1024 ** 3
    return free_space_gb >= required_space_gb


async def is_4xx_client_error(status_code: int) -> bool:
    """Checks whether the HTTP status code is 4xx client error"""
    return HTTPStatus.BAD_REQUEST <= status_code < HTTPStatus.INTERNAL_SERVER_ERROR



import asyncio
from typing import Dict, Optional

import aiofiles
from yarl import URL

from cyberdrop_dl.base_functions.base_functions import log
from cyberdrop_dl.base_functions.data_classes import SkipData, CascadeItem, ForumItem, AlbumItem, DomainItem
from cyberdrop_dl.base_functions.sql_helper import SQLHelper
from cyberdrop_dl.client.client import Client, ScrapeSession
from cyberdrop_dl.client.rate_limiting import AsyncRateLimiter
from cyberdrop_dl.crawlers.Anonfiles_Spider import AnonfilesCrawler
from cyberdrop_dl.crawlers.Bunkr_Spider import BunkrCrawler
from cyberdrop_dl.crawlers.Coomeno_Spider import CoomenoCrawler
from cyberdrop_dl.crawlers.CyberFile_Spider import CyberFileCrawler
from cyberdrop_dl.crawlers.Cyberdrop_Spider import CyberdropCrawler
from cyberdrop_dl.crawlers.EHentai_Spider import EHentaiCrawler
from cyberdrop_dl.crawlers.Erome_Spider import EromeCrawler
from cyberdrop_dl.crawlers.Fapello_Spider import FapelloCrawler
from cyberdrop_dl.crawlers.Gfycat_Spider import GfycatCrawler
from cyberdrop_dl.crawlers.GoFile_Spider import GoFileCrawler
from cyberdrop_dl.crawlers.HGameCG_Spider import HGameCGCrawler
from cyberdrop_dl.crawlers.ImgBox_Spider import ImgBoxCrawler
from cyberdrop_dl.crawlers.LoveFap_Spider import LoveFapCrawler
from cyberdrop_dl.crawlers.NSFWXXXCrawler import NSFWXXXCrawler
from cyberdrop_dl.crawlers.PimpAndHost_Spider import PimpAndHostCrawler
from cyberdrop_dl.crawlers.PixelDrain_Spider import PixelDrainCrawler
from cyberdrop_dl.crawlers.PostImg_Spider import PostImgCrawler
from cyberdrop_dl.crawlers.Saint_Spider import SaintCrawler
from cyberdrop_dl.crawlers.ShareX_Spider import ShareXCrawler
from cyberdrop_dl.crawlers.XBunkr_Spider import XBunkrCrawler
from cyberdrop_dl.crawlers.Xenforo_Spider import XenforoCrawler
from cyberdrop_dl.scraper.JDownloader_Integration import JDownloader


class ScrapeMapper:
    """This class maps links to their respective handlers, or JDownloader if they are unsupported"""
    def __init__(self, args: Dict, client: Client, SQL_Helper: SQLHelper, quiet: bool):
        self.args = args
        self.client = client
        self.SQL_Helper = SQL_Helper
        self.Cascade = CascadeItem({})
        self.Forums = ForumItem({})
        self.skip_data = SkipData(args['Ignore']['skip_hosts'])

        self.unsupported_file = args["Files"]["unsupported_urls_file"]
        self.unsupported_output = args['Runtime']['output_unsupported_urls']

        self.anonfiles_crawler = None
        self.bunkr_crawler = None
        self.cyberdrop_crawler = None
        self.coomeno_crawler = None
        self.cyberfile_crawler = None
        self.ehentai_crawler = None
        self.erome_crawler = None
        self.fapello_crawler = None
        self.gfycat_crawler = None
        self.gofile_crawler = None
        self.hgamecg_crawler = None
        self.imgbox_crawler = None
        self.lovefap_crawler = None
        self.nsfwxxx_crawler = None
        self.pimpandhost_crawler = None
        self.pixeldrain_crawler = None
        self.postimg_crawler = None
        self.redgifs_crawler = None
        self.rule34_crawler = None
        self.saint_crawler = None
        self.sharex_crawler = None
        self.xbunkr_crawler = None
        self.xenforo_crawler = None

        self.include_id = args['Runtime']['include_id']
        self.remove_bunkr_id = args['Runtime']['remove_bunkr_identifier']
        self.separate_posts = args["Forum_Options"]["separate_posts"]
        self.quiet = quiet
        self.jdownloader = JDownloader(args['JDownloader'], quiet)

        self.jpgfish_limiter = AsyncRateLimiter(10)
        self.bunkr_limiter = AsyncRateLimiter(15)
        self.coomeno_limiter = AsyncRateLimiter(8)
        self.gofile_limiter = AsyncRateLimiter(max_calls=1, period=2)

        self.gofile_semaphore = asyncio.Semaphore(1)
        self.jpgfish_semaphore = asyncio.Semaphore(5)
        self.cyberfile_semaphore = asyncio.Semaphore(2)
        self.simpcity_semaphore = asyncio.Semaphore(1)
        self.socialmediagirls_semaphore = asyncio.Semaphore(1)
        self.xbunker_semaphore = asyncio.Semaphore(1)

        self.mapping = {"anonfiles": self.Anonfiles, "bayfiles": self.Anonfiles,"xbunkr": self.XBunkr,
                        "bunkr": self.Bunkr, "cyberdrop": self.Cyberdrop, "cyberfile": self.CyberFile,
                        "erome": self.Erome, "fapello": self.Fapello, "gfycat": self.Gfycat, "gofile": self.GoFile,
                        "hgamecg": self.HGameCG, "imgbox": self.ImgBox, "pixeldrain": self.PixelDrain,
                        "postimg": self.PostImg, "saint": self.Saint, "img.kiwi": self.ShareX,
                        "jpg.church": self.ShareX, "jpg.fish": self.ShareX, "pixl.li": self.ShareX,
                        "nsfw.xxx": self.NSFW_XXX, "pimpandhost": self.PimpAndHost, "lovefap": self.LoveFap,
                        "e-hentai": self.EHentai, "gallery.deltaporno": self.ShareX,
                        "coomer.party": self.Coomeno, "kemono.party": self.Coomeno,
                        "simpcity": self.Xenforo, "socialmediagirls": self.Xenforo, "xbunker": self.Xenforo}

    async def handle_additions(self, domain: str, album_obj: Optional[AlbumItem], domain_obj: Optional[DomainItem], title=None):
        if album_obj:
            if title:
                await album_obj.append_title(title)
                await self.Forums.add_album_to_thread(title, domain, album_obj)
            else:
                await self.Cascade.add_album(domain, album_obj.title, album_obj)
        if domain_obj:
            if title:
                await domain_obj.append_title(title)
                for album_title, album in domain_obj.albums.items():
                    await self.Forums.add_album_to_thread(title, domain, album)
            else:
                for title, album in domain_obj.albums.items():
                    await self.Cascade.add_album(domain, album.title, album)

    """Regular file host handling"""

    async def Anonfiles(self, url: URL, title=None):
        anonfiles_session = ScrapeSession(self.client)
        if not self.anonfiles_crawler:
            self.anonfiles_crawler = AnonfilesCrawler(quiet=self.quiet, SQL_Helper=self.SQL_Helper)
        album_obj = await self.anonfiles_crawler.fetch(anonfiles_session, url)
        await self.handle_additions("anonfiles", album_obj, None, title)
        await anonfiles_session.exit_handler()

    async def Bunkr(self, url: URL, title=None):
        bunkr_session = ScrapeSession(self.client)
        if not self.bunkr_crawler:
            self.bunkr_crawler = BunkrCrawler(quiet=self.quiet, SQL_Helper=self.SQL_Helper,
                                              remove_bunkr_id=self.remove_bunkr_id)
        async with self.bunkr_limiter:
            album_obj = await self.bunkr_crawler.fetch(bunkr_session, url)
        if not await album_obj.is_empty():
            await self.handle_additions("bunkr", album_obj, None, title)
        await bunkr_session.exit_handler()

    async def Cyberdrop(self, url: URL, title=None):
        cyberdrop_session = ScrapeSession(self.client)
        if not self.cyberdrop_crawler:
            self.cyberdrop_crawler = CyberdropCrawler(include_id=self.include_id, quiet=self.quiet, SQL_Helper=self.SQL_Helper)
        album_obj = await self.cyberdrop_crawler.fetch(cyberdrop_session, url)
        await self.handle_additions("cyberdrop", album_obj, None, title)
        await cyberdrop_session.exit_handler()

    async def CyberFile(self, url, title=None):
        cyberfile_session = ScrapeSession(self.client)
        if not self.cyberfile_crawler:
            self.cyberfile_crawler = CyberFileCrawler(quiet=self.quiet, SQL_Helper=self.SQL_Helper)
        async with self.cyberfile_semaphore:
            domain_obj = await self.cyberfile_crawler.fetch(cyberfile_session, url)
        await self.handle_additions("cyberfile", None, domain_obj, title)
        await cyberfile_session.exit_handler()

    async def EHentai(self, url, title=None):
        ehentai_session = ScrapeSession(self.client)
        if not self.ehentai_crawler:
            self.ehentai_crawler = EHentaiCrawler(quiet=self.quiet, SQL_Helper=self.SQL_Helper)
        album_obj = await self.ehentai_crawler.fetch(ehentai_session, url)
        await self.handle_additions("e-hentai", album_obj, None, title)
        await ehentai_session.exit_handler()

    async def Erome(self, url, title=None):
        erome_session = ScrapeSession(self.client)
        if not self.erome_crawler:
            self.erome_crawler = EromeCrawler(include_id=self.include_id, quiet=self.quiet, SQL_Helper=self.SQL_Helper)
        domain_obj = await self.erome_crawler.fetch(erome_session, url)
        await self.handle_additions("erome", None, domain_obj, title)
        await erome_session.exit_handler()

    async def Gfycat(self, url, title=None):
        gfycat_session = ScrapeSession(self.client)
        if not self.gfycat_crawler:
            self.gfycat_crawler = GfycatCrawler(quiet=self.quiet, SQL_Helper=self.SQL_Helper)
        album_obj = await self.gfycat_crawler.fetch(gfycat_session, url)
        await self.handle_additions("gfycat", album_obj, None, title)
        await gfycat_session.exit_handler()

    async def GoFile(self, url, title=None):
        gofile_session = ScrapeSession(self.client)
        if not self.gofile_crawler:
            self.gofile_crawler = GoFileCrawler(quiet=self.quiet, SQL_Helper=self.SQL_Helper)
        async with self.gofile_semaphore:
            await self.gofile_crawler.get_token(session=gofile_session)
        async with self.gofile_limiter:
            domain_obj = await self.gofile_crawler.fetch(gofile_session, url)
        await self.handle_additions("gofile", None, domain_obj, title)
        await gofile_session.exit_handler()

    async def HGameCG(self, url, title=None):
        hgamecg_session = ScrapeSession(self.client)
        if not self.hgamecg_crawler:
            self.hgamecg_crawler = HGameCGCrawler(quiet=self.quiet, SQL_Helper=self.SQL_Helper)
        album_obj = await self.hgamecg_crawler.fetch(hgamecg_session, url)
        await self.handle_additions("hgamecg", album_obj, None, title)
        await hgamecg_session.exit_handler()

    async def ImgBox(self, url, title=None):
        imgbox_session = ScrapeSession(self.client)
        if not self.imgbox_crawler:
            self.imgbox_crawler = ImgBoxCrawler(quiet=self.quiet, SQL_Helper=self.SQL_Helper)
        album_obj = await self.imgbox_crawler.fetch(imgbox_session, url)
        await self.handle_additions("imgbox", album_obj, None, title)
        await imgbox_session.exit_handler()

    async def LoveFap(self, url: URL, title=None):
        lovefap_session = ScrapeSession(self.client)
        if not self.lovefap_crawler:
            self.lovefap_crawler = LoveFapCrawler(quiet=self.quiet, SQL_Helper=self.SQL_Helper)
        album_obj = await self.lovefap_crawler.fetch(lovefap_session, url)
        await self.handle_additions("lovefap", album_obj, None, title)
        await lovefap_session.exit_handler()

    async def PimpAndHost(self, url, title=None):
        pimpandhost_session = ScrapeSession(self.client)
        if not self.pimpandhost_crawler:
            self.pimpandhost_crawler = PimpAndHostCrawler(quiet=self.quiet, SQL_Helper=self.SQL_Helper)
        album_obj = await self.pimpandhost_crawler.fetch(pimpandhost_session, url)
        await self.handle_additions("pimpandhost", album_obj, None, title)
        await pimpandhost_session.exit_handler()

    async def PixelDrain(self, url, title=None):
        pixeldrain_session = ScrapeSession(self.client)
        if not self.pixeldrain_crawler:
            self.pixeldrain_crawler = PixelDrainCrawler(quiet=self.quiet, SQL_Helper=self.SQL_Helper)
        album_obj = await self.pixeldrain_crawler.fetch(pixeldrain_session, url)
        await self.handle_additions("pixeldrain", album_obj, None, title)
        await pixeldrain_session.exit_handler()

    async def PostImg(self, url, title=None):
        postimg_session = ScrapeSession(self.client)
        if not self.postimg_crawler:
            self.postimg_crawler = PostImgCrawler(quiet=self.quiet, SQL_Helper=self.SQL_Helper)
        album_obj = await self.postimg_crawler.fetch(postimg_session, url)
        await self.handle_additions("postimg", album_obj, None, title)
        await postimg_session.exit_handler()

    async def Saint(self, url, title=None):
        saint_session = ScrapeSession(self.client)
        if not self.saint_crawler:
            self.saint_crawler = SaintCrawler(quiet=self.quiet, SQL_Helper=self.SQL_Helper)
        album_obj = await self.saint_crawler.fetch(saint_session, url)
        await self.handle_additions("saint", album_obj, None, title)
        await saint_session.exit_handler()

    async def ShareX(self, url, title=None):
        sharex_session = ScrapeSession(self.client)
        if not self.sharex_crawler:
            self.sharex_crawler = ShareXCrawler(include_id=self.include_id, quiet=self.quiet, SQL_Helper=self.SQL_Helper)
        if ("jpg.fish" in url.host or "jpg.church" in url.host or "img.kiwi" in url.host) and sharex_session.client.ratelimit > 19:
            async with self.jpgfish_semaphore:
                async with self.jpgfish_limiter:
                    domain_obj = await self.sharex_crawler.fetch(sharex_session, url)
        else:
            domain_obj = await self.sharex_crawler.fetch(sharex_session, url)
        await self.handle_additions("sharex", None, domain_obj, title)
        await sharex_session.exit_handler()

    async def XBunkr(self, url, title=None):
        xbunkr_session = ScrapeSession(self.client)
        if not self.xbunkr_crawler:
            self.xbunkr_crawler = XBunkrCrawler(quiet=self.quiet, SQL_Helper=self.SQL_Helper)
        album_obj = await self.xbunkr_crawler.fetch(xbunkr_session, url)
        await self.handle_additions("xbunkr", album_obj, None, title)
        await xbunkr_session.exit_handler()

    """Archive Sites"""

    async def Fapello(self, url, title=None):
        fapello_session = ScrapeSession(self.client)
        if not self.fapello_crawler:
            self.fapello_crawler = FapelloCrawler(quiet=self.quiet, SQL_Helper=self.SQL_Helper)
        album_obj = await self.fapello_crawler.fetch(fapello_session, url)
        await self.handle_additions("fapello", album_obj, None, title)
        await fapello_session.exit_handler()

    async def NSFW_XXX(self, url, title=None):
        nsfwxxx_session = ScrapeSession(self.client)
        if not self.nsfwxxx_crawler:
            self.nsfwxxx_crawler = NSFWXXXCrawler(separate_posts=self.separate_posts, quiet=self.quiet, SQL_Helper=self.SQL_Helper)
        domain_obj = await self.nsfwxxx_crawler.fetch(nsfwxxx_session, url)
        await self.handle_additions("nsfw.xxx", None, domain_obj, title)
        await nsfwxxx_session.exit_handler()

    async def Coomeno(self, url: URL, title=None):
        coomeno_session = ScrapeSession(self.client)
        if not self.coomeno_crawler:
            self.coomeno_crawler = CoomenoCrawler(include_id=self.include_id, scraping_mapper=self,
                                                  separate_posts=self.separate_posts, SQL_Helper=self.SQL_Helper,
                                                  quiet=self.quiet)
        async with self.coomeno_limiter:
            cascade, new_title = await self.coomeno_crawler.fetch(coomeno_session, url)
        if not new_title or await cascade.is_empty():
            await coomeno_session.exit_handler()
            return
        if title:
            await cascade.append_title(title)
            await self.Forums.extend_thread(title, cascade)
        else:
            await self.Forums.add_thread(new_title, cascade)
        await coomeno_session.exit_handler()

    """Forum handling"""

    async def Xenforo(self, url: URL, title=None):
        xenforo_session = ScrapeSession(self.client)
        if not self.xenforo_crawler:
            self.xenforo_crawler = XenforoCrawler(scraping_mapper=self, args=self.args, SQL_Helper=self.SQL_Helper,
                                                  quiet=self.quiet)
        title = None
        cascade = None

        if "simpcity" in url.host:
            async with self.simpcity_semaphore:
                cascade, title = await self.xenforo_crawler.fetch(xenforo_session, url)
        if "socialmediagirls" in url.host:
            async with self.socialmediagirls_semaphore:
                cascade, title = await self.xenforo_crawler.fetch(xenforo_session, url)
        if "xbunker" in url.host:
            async with self.xbunker_semaphore:
                cascade, title = await self.xenforo_crawler.fetch(xenforo_session, url)
        if not title or await cascade.is_empty():
            await xenforo_session.exit_handler()
            return
        await self.Forums.add_thread(title, cascade)
        await xenforo_session.exit_handler()

    """URL to Function Mapper"""

    async def map_url(self, url_to_map: URL, title=None, referer=None):
        if not url_to_map:
            return
        elif not url_to_map.host:
            await log(f"[yellow]Not Supported: {str(url_to_map)}[/yellow]", quiet=self.quiet)
            return
        for key, value in self.mapping.items():
            if key in url_to_map.host:
                if any(site in key for site in self.skip_data.sites):
                    await log(f"[yellow]Skipping: {str(url_to_map)}[/yellow]", quiet=self.quiet)
                else:
                    await value(url=url_to_map, title=title)
                return

        if self.jdownloader.jdownloader_enable:
            async with asyncio.Semaphore(1):
                self.jdownloader.quiet = self.quiet
                if not self.jdownloader.jdownloader_agent:
                    await self.jdownloader.jdownloader_setup()
            await self.jdownloader.direct_unsupported_to_jdownloader(url_to_map, title)

        else:
            await log(f"[yellow]Not Supported: {str(url_to_map)}[/yellow]", quiet=self.quiet)
            if self.unsupported_output:
                title = title.encode("ascii", "ignore")
                title = title.decode().strip()
                async with aiofiles.open(self.unsupported_file, mode='a') as f:
                    await f.write(f"{str(url_to_map)},{str(referer)},{title}\n")



import asyncio

from bs4 import BeautifulSoup
from yarl import URL

from ..base_functions.base_functions import log, logger, make_title_safe, get_db_path, get_filename_and_ext
from ..base_functions.data_classes import CascadeItem, MediaItem
from ..base_functions.sql_helper import SQLHelper
from ..client.client import ScrapeSession


class CoomenoCrawler:
    def __init__(self, *, include_id=False, scraping_mapper, separate_posts=False, quiet: bool, SQL_Helper: SQLHelper):
        self.include_id = include_id
        self.quiet = quiet
        self.scraping_mapper = scraping_mapper
        self.separate_posts = separate_posts
        self.SQL_Helper = SQL_Helper

    async def fetch(self, session: ScrapeSession, url: URL):
        """Director for Coomer/Kemono scraping"""
        await log(f"[green]Starting: {str(url)}[/green]", quiet=self.quiet)
        cascade = CascadeItem({})
        title = None
        try:
            if "coomer" in url.host:
                title = await self.handle_coomer(session, url, cascade)
            elif "kemono" in url.host:
                title = await self.handle_kemono(session, url, cascade)
        except Exception as e:
            logger.debug("Error encountered while handling %s", str(url), exc_info=True)
            await log(f"[red]Error: {str(url)}[/red]", quiet=self.quiet)
            logger.debug(e)

        await self.SQL_Helper.insert_cascade(cascade)
        await log(f"[green]Finished: {str(url)}[/green]", quiet=self.quiet)
        return cascade, title

    async def handle_coomer(self, session: ScrapeSession, url: URL, cascade: CascadeItem):
        """Coomer director function"""
        post_selectors = ['h2[class=post-card__heading] a']
        next_page_selector = 'a[title="Next page"]'
        images_selector = 'a[class="fileThumb"]'
        downloads_selector = 'a[class=post__attachment-link]'
        text_selector = 'div[class=post__content] a'

        title = "Loose Coomer Files"
        if "thumbnail" in url.parts:
            parts = [x for x in url.parts if x not in ("thumbnail", "/")]
            link = URL("https://coomer.party/" + "/".join(parts))

            url_path = await get_db_path(link)
            complete = await self.SQL_Helper.check_complete_singular("coomer", url_path)
            filename, ext = await get_filename_and_ext(link.name)
            media_item = MediaItem(link, url, complete, filename, ext, filename)
            await cascade.add_to_album("coomer", title, media_item)

        elif "data" in url.parts:
            url_path = await get_db_path(url)
            complete = await self.SQL_Helper.check_complete_singular("coomer", url_path)
            filename, ext = await get_filename_and_ext(url.name)
            media_item = MediaItem(url, url, complete, filename, ext, filename)
            await cascade.add_to_album("coomer", title, media_item)

        elif "post" in url.parts:
            title = await self.parse_post(session, url, "coomer", cascade, images_selector, downloads_selector,
                                          text_selector)

        else:
            title = await self.parse_profile(session, url, "coomer", cascade, post_selectors, next_page_selector,
                                             images_selector, downloads_selector, text_selector)

        return title

    async def handle_kemono(self, session: ScrapeSession, url: URL, cascade: CascadeItem):
        """Kemono director function"""
        post_selectors = ['article[class="post-card post-card--preview"] a', 'article[class="post-card"] a']
        next_page_selector = 'a[class=next]'
        images_selector = 'a[class="fileThumb"]'
        downloads_selector = 'a[class=post__attachment-link]'
        text_selector = 'div[class=post__content] a'

        title = "Loose Kemono Files"
        if "thumbnail" in url.parts:
            parts = [x for x in url.parts if x not in ("thumbnail", "/")]
            link = URL("https://kemono.party/" + "/".join(parts))

            url_path = await get_db_path(link)
            complete = await self.SQL_Helper.check_complete_singular("kemono", url_path)
            filename, ext = await get_filename_and_ext(link.name)
            media_item = MediaItem(link, url, complete, filename, ext, filename)
            await cascade.add_to_album("kemono", title, media_item)

        elif "data" in url.parts:
            url_path = await get_db_path(url)
            complete = await self.SQL_Helper.check_complete_singular("kemono", url_path)
            filename, ext = await get_filename_and_ext(url.name)
            media_item = MediaItem(url, url, complete, filename, ext, filename)
            await cascade.add_to_album("kemono", title, media_item)

        elif "post" in url.parts:
            title = await self.parse_post(session, url, "kemono", cascade, images_selector, downloads_selector,
                                          text_selector)

        else:
            title = await self.parse_profile(session, url, "kemono", cascade, post_selectors, next_page_selector,
                                             images_selector, downloads_selector, text_selector)

        return title

    async def map_links(self, text_content: list, title: str, referer: URL):
        """Maps external links to other scrapers"""
        tasks = []
        for content in text_content:
            link = URL(content.get('href'))
            tasks.append(asyncio.create_task(self.scraping_mapper.map_url(link, title, referer)))
        if tasks:
            await asyncio.wait(tasks)

    async def parse_profile(self, session: ScrapeSession, url: URL, domain: str, cascade: CascadeItem,
                            posts_selectors: list, next_page_selector: str, images_selector: str,
                            downloads_selector: str, text_selector: str):
        """Parses profiles with supplied selectors"""
        title = None
        try:
            soup = await session.get_BS4(url)
            title = await make_title_safe(soup.select_one("span[itemprop=name]").get_text())
            title = f"{title} ({url.host})"

            posts = []
            for posts_selector in posts_selectors:
                posts += soup.select(posts_selector)
            for post in posts:
                path = post.get('href')
                if path:
                    post_link = URL("https://" + url.host + path)
                    await self.parse_post(session, post_link, domain, cascade, images_selector, downloads_selector,
                                          text_selector, title)

            next_page = soup.select_one(next_page_selector)
            if next_page:
                next_page = next_page.get('href')
                if next_page:
                    await self.parse_profile(session, URL("https://" + url.host + next_page), domain, cascade,
                                             posts_selectors, next_page_selector, images_selector, downloads_selector,
                                             text_selector)

        except Exception as e:
            logger.debug("Error encountered while handling %s", str(url), exc_info=True)
            await log(f"[red]Error: {str(url)}[/red]", quiet=self.quiet)
            logger.debug(e)

        return title

    async def parse_post(self, session: ScrapeSession, url: URL, domain: str, cascade: CascadeItem,
                         images_selector: str, downloads_selector: str, text_selector: str, title=None):
        """Parses posts with supplied selectors"""
        try:
            url_path = await get_db_path(url)
            text = await self.SQL_Helper.get_blob(url_path)
            if not text:
                text = await session.get_text(url)
                await self.SQL_Helper.insert_blob(text, url_path)
            soup = BeautifulSoup(text, 'html.parser')

            if self.separate_posts:
                if self.include_id:
                    if title:
                        title = title + '/' + str(url.parts[-1]) + " - " + await make_title_safe(soup.select_one("h1[class=post__title]").text.replace('\n', '').replace("..", ""))
                    else:
                        title = await make_title_safe(str(url.parts[-1]) + " - " + soup.select_one("h1[class=post__title]").text.replace('\n', '').replace("..", ""))
                else:
                    if title:
                        title = title + '/' + await make_title_safe(soup.select_one("h1[class=post__title]").text.replace('\n', '').replace("..", ""))
                    else:
                        title = await make_title_safe(soup.select_one("h1[class=post__title]").text.replace('\n', '').replace("..", ""))
            else:
                if not title:
                    title = await make_title_safe(soup.select_one("h1[class=post__title]").text.replace('\n', '').replace("..", ""))

            images = soup.select(images_selector)
            for image in images:
                href = image.get('href')
                if href.startswith("/"):
                    link = URL("https://" + url.host + href)
                else:
                    link = URL(href)
                url_path = await get_db_path(link)
                complete = await self.SQL_Helper.check_complete_singular(domain, url_path)
                filename, ext = await get_filename_and_ext(link.name)
                media_item = MediaItem(link, url, complete, filename, ext, filename)
                await cascade.add_to_album(domain, title, media_item)

            downloads = soup.select(downloads_selector)
            for download in downloads:
                href = download.get('href')
                if href.startswith("/"):
                    link = URL("https://" + url.host + href)
                else:
                    link = URL(href)
                url_path = await get_db_path(link)
                complete = await self.SQL_Helper.check_complete_singular(domain, url_path)
                filename, ext = await get_filename_and_ext(link.name)
                media_item = MediaItem(link, url, complete, filename, ext, filename)
                await cascade.add_to_album(domain, title, media_item)

            text_content = soup.select(text_selector)
            await self.map_links(text_content, title, url)
        except Exception as e:
            logger.debug("Error encountered while handling %s", str(url), exc_info=True)
            await log(f"[red]Error: {str(url)}[/red]", quiet=self.quiet)
            logger.debug(e)

        return title



from yarl import URL

from ..base_functions.base_functions import log, logger, make_title_safe, get_filename_and_ext, get_db_path
from ..base_functions.data_classes import AlbumItem, MediaItem
from ..base_functions.error_classes import NoExtensionFailure
from ..base_functions.sql_helper import SQLHelper
from ..client.client import ScrapeSession


class FapelloCrawler:
    def __init__(self, *, quiet: bool, SQL_Helper: SQLHelper):
        self.quiet = quiet
        self.SQL_Helper = SQL_Helper

    async def fetch(self, session: ScrapeSession, url: URL):
        """Basic director for fapello"""
        await log(f"[green]Starting: {str(url)}[/green]", quiet=self.quiet)

        album_obj = await self.parse_profile(session, url)

        await log(f"[green]Finished: {str(url)}[/green]", quiet=self.quiet)
        if not album_obj:
            return None
        return album_obj

    async def parse_profile(self, session: ScrapeSession, url: URL):
        """Profile parser, passes posts to parse_post"""
        try:
            soup, returned_url = await session.get_BS4_and_url(url)
            title = soup.select_one('h2[class="font-semibold lg:text-2xl text-lg mb-2 mt-4"]').get_text()
            title = title + " (Fapello)"
            title = await make_title_safe(title)

            album_obj = AlbumItem(title, [])

            # Fapello does circular looped paging
            if returned_url != url:
                return album_obj

            content_section = soup.select_one("div[id=content]")
            posts = content_section.select("a")
            for post in posts:
                link = post.get('href')
                if link:
                    media_items = await self.parse_post(session, link)
                    for media in media_items:
                        await album_obj.add_media(media)

            next_page = soup.select_one('div[id="next_page"] a')
            if next_page:
                next_page = next_page.get('href')
                if next_page:
                    await album_obj.extend(await self.parse_profile(session, URL(next_page)))
            await self.SQL_Helper.insert_album("fapello", url.path, album_obj)
            return album_obj

        except Exception as e:
            logger.debug("Error encountered while handling %s", str(url), exc_info=True)
            await log(f"[red]Error: {str(url)}[/red]", quiet=self.quiet)
            logger.debug(e)
            return

    async def parse_post(self, session: ScrapeSession, url: URL):
        """Parses posts, returns list of media_items"""
        try:
            soup = await session.get_BS4(url)
            results = []

            content_section = soup.select_one('div[class="flex justify-between items-center"]')

            images = content_section.select("img")
            for image in images:
                download_link = URL(image.get('src'))
                try:
                    filename, ext = await get_filename_and_ext(download_link.name)
                except NoExtensionFailure:
                    logger.debug("Couldn't get extension for %s", str(download_link))
                    continue
                url_path = await get_db_path(download_link)
                complete = await self.SQL_Helper.check_complete_singular("fapello", url_path)
                results.append(MediaItem(download_link, url, complete, filename, ext, filename))

            videos = content_section.select("source")
            for video in videos:
                download_link = URL(video.get('src'))
                try:
                    filename, ext = await get_filename_and_ext(download_link.name)
                except NoExtensionFailure:
                    logger.debug("Couldn't get extension for %s", str(download_link))
                    continue
                url_path = await get_db_path(download_link)
                complete = await self.SQL_Helper.check_complete_singular("fapello", url_path)
                results.append(MediaItem(download_link, url, complete, filename, ext, filename))

            return results

        except Exception as e:
            logger.debug("Error encountered while handling %s", str(url), exc_info=True)
            await log(f"[red]Error: {str(url)}[/red]", quiet=self.quiet)
            logger.debug(e)
            return



from __future__ import annotations

import asyncio
import re
from pathlib import Path

import aiofiles
from bs4 import BeautifulSoup
from yarl import URL

from ..base_functions.base_functions import (
    get_filename_and_ext,
    log,
    logger,
    make_title_safe,
)
from ..base_functions.data_classes import CascadeItem, MediaItem
from ..base_functions.error_classes import FailedLoginFailure, NoExtensionFailure
from ..base_functions.sql_helper import SQLHelper
from ..client.client import ScrapeSession


async def write_last_post_file(file: Path, url: str):
    """Writes the last post url from a thread to the specified file"""
    async with aiofiles.open(file, mode='a') as f:
        await f.write(url + '\n')


class ForumLogin:
    def __init__(self, name: str, username: str, password: str):
        self.name = name
        self.logged_in = False
        self.username = username
        self.password = password

    async def login(self, session: ScrapeSession, url: URL, quiet: bool):
        """Handles forum logging in"""
        if not self.username or not self.password:
            await log(f"[red]Login wasn't provided for {self.name}[/red]", quiet=quiet)
            raise FailedLoginFailure()
        attempt = 0
        while True:
            try:
                if self.logged_in:
                    return
                if attempt == 5:
                    raise FailedLoginFailure()

                domain = URL("https://" + url.host) / "login"
                text = await session.get_text(domain)
                await asyncio.sleep(5)
                soup = BeautifulSoup(text, 'html.parser')

                inputs = soup.select('form input')
                data = {
                    elem['name']: elem['value']
                    for elem in inputs
                    if elem.get('name') and elem.get('value')
                }
                data.update({
                    "login": self.username,
                    "password": self.password,
                    "_xfRedirect": str(URL("https://" + url.host))
                })
                await session.post_data_no_resp(domain / "login", data=data)
                await asyncio.sleep(5)
                text = await session.get_text(domain)
                if "You are already logged in" not in text:
                    raise FailedLoginFailure()

                self.logged_in = True
            except asyncio.exceptions.TimeoutError:
                attempt += 1
                continue


class XenforoCrawler:
    def __init__(self, *, scraping_mapper, args: dict, SQL_Helper: SQLHelper, quiet: bool):
        self.include_id = args["Runtime"]["include_id"]
        self.quiet = quiet
        self.separate_posts = args["Forum_Options"]["separate_posts"]
        self.output_last = args["Forum_Options"]["output_last_forum_post"]
        self.output_last_file = args["Files"]["output_last_forum_post_file"]

        self.simpcity = ForumLogin("SimpCity",
                                   args["Authentication"]["simpcity_username"],
                                   args["Authentication"]["simpcity_password"])

        self.socialmediagirls = ForumLogin("SocialMediaGirls",
                                           args["Authentication"]["socialmediagirls_username"],
                                           args["Authentication"]["socialmediagirls_password"])

        self.xbunker = ForumLogin("XBunker",
                                  args["Authentication"]["xbunker_username"],
                                  args["Authentication"]["xbunker_password"])

        self.scraping_mapper = scraping_mapper
        self.SQL_Helper = SQL_Helper

    async def fetch(self, session: ScrapeSession, url: URL):
        """Xenforo forum director"""
        await log(f"[green]Starting: {str(url)}[/green]", quiet=self.quiet)
        cascade = CascadeItem({})

        scrape_url, post_num = await self.get_thread_url_and_post_num(url)
        title = None
        try:
            if "simpcity" in url.host:
                await self.simpcity.login(session, url, self.quiet)
                title = await self.parse_simpcity(session, scrape_url, cascade, "", post_num)
            elif "socialmediagirls" in url.host:
                await self.socialmediagirls.login(session, url, self.quiet)
                title = await self.parse_socialmediagirls(session, scrape_url, cascade, "", post_num)
            elif "xbunker" in url.host:
                await self.xbunker.login(session, url, self.quiet)
                title = await self.parse_xbunker(session, scrape_url, cascade, "", post_num)

        except Exception as e:
            logger.debug("Error encountered while handling %s", str(url), exc_info=True)
            await log(f"[red]Error: {str(url)}[/red]", quiet=self.quiet)
            logger.debug(e)
            return cascade, ""

        await self.SQL_Helper.insert_cascade(cascade)
        await log(f"[green]Finished: {str(url)}[/green]", quiet=self.quiet)
        return cascade, title

    async def get_thread_url_and_post_num(self, url: URL):
        """Splits the thread url and returns the url and post number if provided"""
        post_number = 0
        if "post-" in str(url):
            post_number_parts = str(url).rsplit("post-", 1)
            post_number = int(post_number_parts[-1].strip("/")) if len(post_number_parts) == 2 else 0
            url = URL(post_number_parts[0].rstrip("#"))
        return url, post_number

    async def get_links(self, post_content, selector, attribute, domain, temp_title):
        """Grabs links from the post content based on the given selector and attribute"""
        found_links = []
        if not post_content:
            return found_links
        links = post_content.select(selector)
        for link in links:
            link = link.get(attribute)
            if link.endswith("/"):
                link = link[:-1]
            if link.startswith('//'):
                link = "https:" + link
            elif link.startswith('/'):
                link = domain / link[1:]
            found_links.append([URL(link), temp_title])
        return found_links

    async def get_embedded(self, post_content, selector, attribute, domain, temp_title):
        """Gets embedded media from post content based on selector and attribute provided"""
        found_links = []
        links = post_content.select(selector)
        for link in links:
            embed_data = link.get(attribute)
            embed_data = embed_data.replace("\/\/", "https://www.")
            embed_data = embed_data.replace("\\", "")

            embed_url = re.search(
                r"https?:\/\/(www\.)?[-a-zA-Z0-9@:%._\+~#=]{2,256}\.[a-z]{2,4}\\b([-a-zA-Z0-9@:%_\+.~#?&//=]*)",
                embed_data)
            if embed_url:
                embed_url = URL(embed_url.group(0).replace("www.", ""))
                found_links.append([embed_url, temp_title])

            embed_url = re.search(
                r"https?:\/\/(www\.)?[-a-zA-Z0-9@:%._\+~#=]{2,256}\.[a-z]{2,4}\/[-a-zA-Z0-9@:%._\+~#=]*\/[-a-zA-Z0-9@:?&%._\+~#=]*",
                embed_data)
            if embed_url:
                embed_url = URL(embed_url.group(0).replace("www.", ""))
                found_links.append([embed_url, temp_title])
        return found_links

    async def filter_content_links(self, cascade: CascadeItem, content_links: list, url: URL, domain: str):
        """Splits given links into direct links and external links,
        returns external links, adds internal to the cascade"""
        forum_direct_urls = [x for x in content_links if x[0].host.replace(".st", ".su") in url.host]
        forum_direct_urls.extend([x for x in content_links if url.host in x[0].host.replace(".st", ".su")])
        forum_direct_urls.extend([x for x in content_links if "smgmedia" in x[0].host])
        content_links = [x for x in content_links if x not in forum_direct_urls]
        for link_title_bundle in forum_direct_urls:
            link = link_title_bundle[0]
            temp_title = link_title_bundle[1]
            in_prog_title = temp_title + "/Attachments"
            if str(link).endswith("/"):
                link = URL(str(link)[:-1])
            if 'attachments' in link.parts or 'content' in link.parts or 'data' in link.parts:
                completed = await self.SQL_Helper.check_complete_singular(domain, link.path)
                try:
                    filename, ext = await get_filename_and_ext(link.name, True)
                except NoExtensionFailure:
                    continue
                media = MediaItem(link, url, completed, filename, ext, filename)
                await cascade.add_to_album(domain, in_prog_title, media)
        return content_links

    async def handle_external_links(self, content_links: list, referer: URL):
        """Maps external links to the scraper class"""
        tasks = []
        for link_title_bundle in content_links:
            link = link_title_bundle[0]
            temp_title = link_title_bundle[1]
            tasks.append(asyncio.create_task(self.scraping_mapper.map_url(link, temp_title, referer)))
        if tasks:
            await asyncio.wait(tasks)

    async def parse_simpcity(self, session: ScrapeSession, url: URL, cascade: CascadeItem, title: str,
                             post_number: int):
        """Parses simpcity threads"""
        soup = await session.get_BS4(url)

        domain = URL("https://" + url.host)
        post_num_str = None
        content_links = []

        title_block = soup.select_one("h1[class=p-title-value]")
        for elem in title_block.find_all("a"):
            elem.decompose()

        if title:
            pass
        else:
            title = title_block.text
            title = await make_title_safe(title.replace("\n", "").strip())

        posts = soup.select("div[class='message-main uix_messageContent js-quickEditTarget']")

        for post in posts:
            post_num_str = post.select_one("li[class=u-concealed] a").get('href').split('/')[-1]
            post_num_int = int(post_num_str.split('post-')[-1])
            if post_number:
                if post_number > post_num_int:
                    continue

            temp_title = title + "/" + post_num_str if self.separate_posts else title

            for elem in post.find_all('blockquote'):
                elem.decompose()
            post_content = post.select_one("div[class=bbWrapper]")

            # Get Links
            content_links.extend(await self.get_links(post_content, "a", "href", domain, temp_title))

            # Get Images
            content_links.extend(await self.get_links(post_content, "div[class='bbImageWrapper js-lbImage']", "data-src", domain, temp_title))
            content_links.extend(await self.get_links(post_content, "div[class='bbImageWrapper lazyload js-lbImage']", "data-src", domain, temp_title))

            # Get Videos:
            content_links.extend(await self.get_links(post_content, "video source", "src", domain, temp_title))
            content_links.extend(await self.get_links(post_content, "iframe[class=saint-iframe]", "src", domain, temp_title))

            # Get Other Embedded Content
            content_links.extend(await self.get_embedded(post_content, "span[data-s9e-mediaembed-iframe]", "data-s9e-mediaembed-iframe",
                                 domain, temp_title))

            # Get Attachments
            attachments_block = post.select_one("section[class=message-attachments]")
            content_links.extend(await self.get_links(attachments_block, "a[class='file-preview js-lbImage']", "href", domain, temp_title))

        # Handle links
        content_links = await self.filter_content_links(cascade, content_links, url, "simpcity")
        await self.handle_external_links(content_links, url)

        next_page = soup.select_one('a[class="pageNav-jump pageNav-jump--next"]')
        if next_page is not None:
            next_page = next_page.get('href')
            if next_page is not None:
                if next_page.startswith('/'):
                    next_page = domain / next_page[1:]
                next_page = URL(next_page)
                await self.parse_simpcity(session, next_page, cascade, title, post_number)
        else:
            if self.output_last:
                if 'page-' in url.raw_name:
                    last_post_url = url.parent / post_num_str
                elif 'post-' in url.raw_name:
                    last_post_url = url.parent / post_num_str
                else:
                    last_post_url = url / post_num_str
                await write_last_post_file(self.output_last_file, str(last_post_url))
        return title

    async def parse_socialmediagirls(self, session: ScrapeSession, url: URL, cascade: CascadeItem, title: str,
                                     post_number: int):
        """Parses SMG threads"""
        soup = await session.get_BS4(url)

        domain = URL("https://" + url.host)
        post_num_str = None
        content_links = []

        title_block = soup.select_one("h1[class=p-title-value]")
        for elem in title_block.find_all("span"):
            elem.decompose()

        if title:
            pass
        else:
            title = title_block.text
            title = await make_title_safe(title.replace("\n", "").strip())

        posts = soup.select("div[class='message-main uix_messageContent js-quickEditTarget']")

        for post in posts:
            post_num_str = post.select_one("li[class=u-concealed] a").get('href').split('/')[-1]
            post_num_int = int(post_num_str.split('post-')[-1])
            if post_number:
                if post_number > post_num_int:
                    continue

            temp_title = title + "/" + post_num_str if self.separate_posts else title

            for elem in post.find_all('blockquote'):
                elem.decompose()
            post_content = post.select_one("div[class=bbWrapper]")

            # Get Links
            content_links.extend(await self.get_links(post_content, "a", "href", domain, temp_title))

            # Get Images
            content_links.extend(await self.get_links(post_content, "div[class='bbImageWrapper js-lbImage']", "data-src", domain, temp_title))
            content_links.extend(await self.get_links(post_content, "div[class='bbImageWrapper lazyload js-lbImage']", "data-src", domain, temp_title))

            # Get Videos:
            content_links.extend(await self.get_links(post_content, "video source", "src", domain, temp_title))
            content_links.extend(await self.get_links(post_content, "iframe[class=saint-iframe]", "src", domain, temp_title))

            # Get Other Embedded Content
            content_links.extend(await self.get_embedded(post_content, "span[data-s9e-mediaembed-iframe]", "data-s9e-mediaembed-iframe", domain, temp_title))

            # Get Attachments
            attachments_block = post.select_one("section[class=message-attachments]")
            content_links.extend(await self.get_links(attachments_block, "a[class='file-preview js-lbImage']", "href", domain, temp_title))

        # Handle links
        content_links = await self.filter_content_links(cascade, content_links, url, "socialmediagirls")
        await self.handle_external_links(content_links, url)

        next_page = soup.select_one('a[class="pageNav-jump pageNav-jump--next"]')
        if next_page is not None:
            next_page = next_page.get('href')
            if next_page is not None:
                if next_page.startswith('/'):
                    next_page = domain / next_page[1:]
                next_page = URL(next_page)
                await self.parse_socialmediagirls(session, next_page, cascade, title, post_number)
        else:
            if self.output_last:
                if 'page-' in url.raw_name:
                    last_post_url = url.parent / post_num_str
                elif 'post-' in url.raw_name:
                    last_post_url = url.parent / post_num_str
                else:
                    last_post_url = url / post_num_str
                await write_last_post_file(self.output_last_file, str(last_post_url))
        return title

    async def parse_xbunker(self, session: ScrapeSession, url: URL, cascade: CascadeItem, title: str,
                            post_number: int):
        """Parses XBunker threads"""
        soup = await session.get_BS4(url)

        domain = URL("https://" + url.host)
        post_num_str = None
        content_links = []

        title_block = soup.select_one("h1[class=p-title-value]")
        for elem in title_block.find_all("a"):
            elem.decompose()

        if title:
            pass
        else:
            title = title_block.text
            title = await make_title_safe(title.replace("\n", "").strip())

        posts = soup.select("div[class='message-main uix_messageContent js-quickEditTarget']")

        for post in posts:
            post_num_str = post.select_one("li[class=u-concealed] a").get('href').split('/')[-1]
            post_num_int = int(post_num_str.split('post-')[-1])
            if post_number:
                if post_number > post_num_int:
                    continue

            temp_title = title + "/" + post_num_str if self.separate_posts else title

            for elem in post.find_all('blockquote'):
                elem.decompose()
            post_content = post.select_one("div[class=bbWrapper]")

            # Get Links
            content_links.extend(await self.get_links(post_content, "a", "href", domain, temp_title))

            # Get Images
            content_links.extend(await self.get_links(post_content, "div[class='bbImageWrapper js-lbImage']", "data-src", domain, temp_title))
            content_links.extend(await self.get_links(post_content, "div[class='bbImageWrapper lazyload js-lbImage']", "data-src", domain, temp_title))

            # Get Videos:
            content_links.extend(await self.get_links(post_content, "video source", "src", domain, temp_title))
            content_links.extend(await self.get_links(post_content, "iframe[class=saint-iframe]", "src", domain, temp_title))

            # Get Other Embedded Content
            content_links.extend(await self.get_embedded(post_content, "span[data-s9e-mediaembed-iframe]", "data-s9e-mediaembed-iframe", domain, temp_title))

            # Get Attachments
            attachments_block = post.select_one("section[class=message-attachments]")
            content_links.extend(await self.get_links(attachments_block, "a[class='file-preview js-lbImage']", "href", domain, temp_title))

        # Handle links
        content_links = await self.filter_content_links(cascade, content_links, url, "xbunker")
        await self.handle_external_links(content_links, url)

        next_page = soup.select_one('a[class="pageNav-jump pageNav-jump--next"]')
        if next_page is not None:
            next_page = next_page.get('href')
            if next_page is not None:
                if next_page.startswith('/'):
                    next_page = domain / next_page[1:]
                next_page = URL(next_page)
                await self.parse_xbunker(session, next_page, cascade, title, post_number)
        else:
            if self.output_last:
                if 'page-' in url.raw_name:
                    last_post_url = url.parent / post_num_str
                elif 'post-' in url.raw_name:
                    last_post_url = url.parent / post_num_str
                else:
                    last_post_url = url / post_num_str
                await write_last_post_file(self.output_last_file, str(last_post_url))
        return title



from __future__ import annotations

import asyncio
import logging
import multiprocessing
from functools import wraps
from http import HTTPStatus
from random import gauss

import aiofiles
import aiohttp.client_exceptions
from rich.live import Live
from rich.progress import TaskID
from yarl import URL

from cyberdrop_dl.base_functions.base_functions import (
    clear,
    get_db_path,
    log,
    logger,
)
from cyberdrop_dl.base_functions.data_classes import AlbumItem, CascadeItem, DomainItem, FileLock, ForumItem, MediaItem
from cyberdrop_dl.base_functions.error_classes import DownloadFailure
from cyberdrop_dl.base_functions.sql_helper import SQLHelper
from cyberdrop_dl.client.client import Client, DownloadSession
from cyberdrop_dl.scraper.Scraper import ScrapeMapper

from .downloader_utils import (
    allowed_filetype,
    basic_auth,
    check_free_space,
    is_4xx_client_error,
)
from .progress_definitions import (
    album_progress,
    cascade_progress,
    domain_progress,
    file_progress,
    forum_progress,
    get_cascade_table,
    get_forum_table,
    overall_file_progress,
)


def retry(f):
    """This function is a wrapper that handles retrying for failed downloads"""

    @wraps(f)
    async def wrapper(self, *args, **kwargs):
        while True:
            try:
                return await f(self, *args, **kwargs)
            except DownloadFailure as e:
                if not self.disable_attempt_limit:
                    if self.current_attempt[args[3]] >= self.allowed_attempts - 1:
                        await self.output_failed(args[2], e)
                        logger.debug('Skipping %s...', args[2].url, exc_info=True)
                        overall_file_progress.advance(self.files.failed_files_task_id, 1)
                        self.files.failed_files += 1
                        return
                logger.debug(e.message)
                logger.debug(f'Retrying ({self.current_attempt[args[3]]}) {args[2].url}...')
                self.current_attempt[args[3]] += 1

    return wrapper


class Files:
    """Class that keeps track of completed, skipped and failed files"""

    def __init__(self, completed, skipped, failed):
        self.completed_files_task_id = completed
        self.completed_files = 0
        self.skipped_files_task_id = skipped
        self.skipped_files = 0
        self.failed_files_task_id = failed
        self.failed_files = 0

    async def hide(self):
        overall_file_progress.update(self.completed_files_task_id, visible=False)
        overall_file_progress.update(self.skipped_files_task_id, visible=False)
        overall_file_progress.update(self.failed_files_task_id, visible=False)


class Downloader:
    """Downloader class, directs downloading for domain objects"""

    def __init__(self, args: dict, client: Client, SQL_Helper: SQLHelper, scraper: ScrapeMapper, max_workers: int,
                 domain: str, domain_obj: DomainItem, semaphore: asyncio.Semaphore, files: Files):
        self.backup_scraper = scraper
        self.client = client
        self.download_session = DownloadSession(client)
        self.File_Lock = FileLock()
        self.SQL_Helper = SQL_Helper

        self.domain = domain
        self.domain_obj = domain_obj

        self.errored_output = args['Runtime']['output_errored_urls']
        self.errored_file = args['Files']['errored_urls_file']

        self.files = files

        self.current_attempt = {}
        self.max_workers = max_workers
        self._semaphore = semaphore
        self.delay = {'cyberfile': 1, 'anonfiles': 1, "coomer": 0.2, "kemono": 0.2}

        self.pixeldrain_api_key = args["Authentication"]["pixeldrain_api_key"]

        self.ignore_history = args["Ignore"]["ignore_history"]
        self.exclude_audio = args["Ignore"]["exclude_audio"]
        self.exclude_images = args["Ignore"]["exclude_images"]
        self.exclude_videos = args["Ignore"]["exclude_videos"]
        self.exclude_other = args["Ignore"]["exclude_other"]

        self.block_sub_folders = args['Runtime']['block_sub_folders']
        self.allowed_attempts = args["Runtime"]["attempts"]
        self.allow_insecure_connections = args["Runtime"]["allow_insecure_connections"]
        self.disable_attempt_limit = args["Runtime"]["disable_attempt_limit"]
        self.download_dir = args["Files"]["output_folder"]
        self.mark_downloaded = args["Runtime"]["skip_download_mark_completed"]
        self.proxy = args["Runtime"]["proxy"]
        self.required_free_space = args["Runtime"]["required_free_space"]

    async def start_domain(self, cascade_task: TaskID):
        """Handler for domains and the progress bars for it"""
        domain_task = domain_progress.add_task("[light_pink3]" + self.domain.upper(), total=len(self.domain_obj.albums))
        for album, album_obj in self.domain_obj.albums.items():
            await self.start_album(domain_task, album, album_obj)
        cascade_progress.advance(cascade_task, 1)
        domain_progress.update(domain_task, visible=False)
        await self.download_session.exit_handler()

    async def start_album(self, domain_task: TaskID, album: str, album_obj: AlbumItem):
        """Handler for albums and the progress bars for it"""
        if await album_obj.is_empty():
            return
        task_description = album.split('/')[-1]
        task_description = task_description.encode("ascii", "ignore").decode().strip()
        if len(task_description) >= 40:
            task_description = task_description[:37] + "..."
        else:
            task_description = task_description.ljust(40)
        album_task = album_progress.add_task("[pink3]" + task_description.upper(), total=len(album_obj.media))
        download_tasks = []
        for media in album_obj.media:
            download_tasks.append(self.start_file(album_task, album, media))
        await asyncio.gather(*download_tasks)
        album_progress.update(album_task, visible=False)
        domain_progress.advance(domain_task, 1)

    async def start_file(self, album_task: TaskID, album: str, media: MediaItem):
        """Handler for files and the progress bars for it"""
        media.original_filename = media.filename

        if media.complete:
            await log(f"Previously Downloaded: {media.filename}", quiet=True)
            overall_file_progress.advance(self.files.skipped_files_task_id, 1)
            self.files.skipped_files += 1
            album_progress.advance(album_task, 1)
            return
        else:
            url_path = await get_db_path(URL(media.url), self.domain)
            complete = await self.SQL_Helper.check_complete_singular(self.domain, url_path)
            if complete:
                await log(f"Previously Downloaded: {media.filename}", quiet=True)
                overall_file_progress.advance(self.files.skipped_files_task_id, 1)
                self.files.skipped_files += 1
                album_progress.advance(album_task, 1)
                return
        async with self._semaphore:
            url_path = await get_db_path(URL(media.url), self.domain)
            await self.download_file(album_task, album, media, url_path)

    @retry
    async def download_file(self, album_task: TaskID, album: str, media: MediaItem, url_path: str):
        """File downloader"""
        if not await check_free_space(self.required_free_space, self.download_dir):
            await log("We've run out of free space.", quiet=True)
            overall_file_progress.advance(self.files.skipped_files_task_id, 1)
            self.files.skipped_files += 1
            album_progress.advance(album_task, 1)
            return

        if not await allowed_filetype(media, self.exclude_images, self.exclude_videos, self.exclude_audio,
                                      self.exclude_other):
            await log(f"Blocked by file extension: {media.filename}", quiet=True)
            overall_file_progress.advance(self.files.skipped_files_task_id, 1)
            self.files.skipped_files += 1
            album_progress.advance(album_task, 1)
            return

        if self.block_sub_folders:
            album = album.split('/')[0]

        original_filename = media.original_filename

        try:
            while await self.File_Lock.check_lock(original_filename):
                await asyncio.sleep(gauss(1, 1.5))
            await self.File_Lock.add_lock(original_filename)

            if url_path not in self.current_attempt:
                self.current_attempt[url_path] = 0

            current_throttle = self.client.throttle

            complete_file = (self.download_dir / album / media.filename)
            partial_file = complete_file.with_suffix(complete_file.suffix + '.part')

            fake_download = False
            if self.mark_downloaded:
                fake_download = True

            complete_file, partial_file, proceed = await self.check_file_exists(complete_file, partial_file, media,
                                                                                album, url_path, original_filename,
                                                                                current_throttle)
            if not proceed:
                fake_download = True

            await self.SQL_Helper.update_pre_download(complete_file, media.filename, url_path, original_filename)

            resume_point = 0
            await self.SQL_Helper.sql_insert_temp(str(partial_file))
            range_num = None
            if partial_file.exists():
                resume_point = partial_file.stat().st_size
                range_num = f'bytes={resume_point}-'

            for key, value in self.delay.items():
                if key in media.url.host:
                    current_throttle = value

            headers = {"Authorization": await basic_auth("Cyberdrop-DL", self.pixeldrain_api_key)} \
                if (self.pixeldrain_api_key and "pixeldrain" in media.url.host) else {}
            if range_num:
                headers['Range'] = range_num

            task_description = media.filename
            task_description = task_description.encode("ascii", "ignore").decode().strip()
            if len(task_description) >= 40:
                task_description = task_description[:37] + "..."
            else:
                task_description = task_description.ljust(40)
            file_task = file_progress.add_task("[plum3]" + task_description, progress_type="file")

            if not await self.SQL_Helper.sql_check_old_existing(url_path) and not fake_download:
                await self.download_session.download_file(media, partial_file, current_throttle, resume_point,
                                                          self.proxy, headers, file_task)
                partial_file.rename(complete_file)

            await self.SQL_Helper.mark_complete(url_path, original_filename)
            if media.url.parts[-1] in self.current_attempt.keys():
                self.current_attempt.pop(media.url.parts[-1])

            if fake_download:
                overall_file_progress.advance(self.files.skipped_files_task_id, 1)
                await log(f"Already Downloaded: {media.filename} from {media.referer}", quiet=True)
                self.files.skipped_files += 1
            else:
                overall_file_progress.advance(self.files.completed_files_task_id, 1)
                self.files.completed_files += 1
            album_progress.advance(album_task, 1)
            file_progress.update(file_task, visible=False)

            await log(f"Completed Download: {media.filename} from {media.referer}", quiet=True)
            await self.File_Lock.remove_lock(original_filename)
            return

        except (aiohttp.client_exceptions.ClientPayloadError, aiohttp.client_exceptions.ClientOSError,
                aiohttp.client_exceptions.ServerDisconnectedError, asyncio.TimeoutError,
                aiohttp.client_exceptions.ClientResponseError, aiohttp.client_exceptions.ServerTimeoutError,
                DownloadFailure, FileNotFoundError, PermissionError) as e:
            if await self.File_Lock.check_lock(original_filename):
                await self.File_Lock.remove_lock(original_filename)

            new_error = DownloadFailure(code=1)
            try:
                file_progress.update(file_task, visible=False)
            except Exception:
                pass

            if hasattr(e, "message"):
                logging.debug(f"\n{media.url} ({e.message})")
            new_error.message = repr(e)

            if hasattr(e, "code"):
                if await is_4xx_client_error(e.code) and e.code != HTTPStatus.TOO_MANY_REQUESTS:
                    logger.debug("We ran into a 400 level error: %s", str(e.code))
                    await log(f"Failed Download: {media.filename}", quiet=True)
                    overall_file_progress.advance(self.files.failed_files_task_id, 1)
                    self.files.failed_files += 1
                    if url_path in self.current_attempt.keys():
                        self.current_attempt.pop(url_path)
                    await self.output_failed(media, e)
                    return
                if e.code == HTTPStatus.SERVICE_UNAVAILABLE or e.code == 521:
                    if hasattr(e, "message"):
                        if not e.message:
                            e.message = "Web server is down"
                        logging.debug(f"\n{media.url} ({e.message})")
                    await log(f"Failed Download: {media.filename}", quiet=True)
                    overall_file_progress.advance(self.files.failed_files_task_id, 1)
                    self.files.failed_files += 1
                    if url_path in self.current_attempt.keys():
                        self.current_attempt.pop(url_path)
                    await self.output_failed(media, e)
                    return

                logger.debug("Error status code: " + str(e.code))
                new_error.code = e.code

            raise new_error

    async def output_failed(self, media, e):
        if self.errored_output:
            async with aiofiles.open(self.errored_file, mode='a') as file:
                await file.write(f"{media.url},{media.referer},{e.message}\n")

    async def check_file_exists(self, complete_file, partial_file, media, album, url_path, original_filename,
                                current_throttle):
        """Complicated checker for if a file already exists, and was already downloaded"""
        expected_size = None
        proceed = True
        while True:
            if complete_file.exists() or partial_file.exists():
                if not expected_size:
                    expected_size = await self.download_session.get_filesize(media.url, str(media.referer),
                                                                             current_throttle)
                if complete_file.exists():
                    if complete_file.stat().st_size == expected_size:
                        proceed = False
                        break
                downloaded_filename = await self.SQL_Helper.get_downloaded_filename(url_path, original_filename)
                if downloaded_filename:
                    if media.filename == downloaded_filename:
                        if partial_file.exists():
                            if partial_file.stat().st_size == expected_size:
                                proceed = False
                                partial_file.rename(complete_file)
                                break
                            else:
                                break
                        elif complete_file.exists():
                            if complete_file.stat().st_size == expected_size:
                                proceed = False
                                break
                            else:
                                complete_file, partial_file = await self.iterate_filename(complete_file, media, album)
                                break
                        else:
                            break
                    else:
                        media.filename = downloaded_filename
                        complete_file = (self.download_dir / album / media.filename)
                        partial_file = complete_file.with_suffix(complete_file.suffix + '.part')
                        continue
                else:
                    complete_file, partial_file = await self.iterate_filename(complete_file, media, album)
                    break
            else:
                break
        return complete_file, partial_file, proceed

    async def iterate_filename(self, complete_file, media, album):
        iterations = 1
        while True:
            filename = f"{complete_file.stem} ({iterations}){media.ext}"
            iterations += 1
            temp_complete_file = (self.download_dir / album / filename)
            if not temp_complete_file.exists():
                if not await self.SQL_Helper.check_filename(filename):
                    media.filename = filename
                    complete_file = (self.download_dir / album / media.filename)
                    partial_file = complete_file.with_suffix(complete_file.suffix + '.part')
                    break
        return complete_file, partial_file


async def download_cascade(args: dict, Cascade: CascadeItem, SQL_Helper: SQLHelper, client: Client,
                           scraper: ScrapeMapper) -> None:
    """Handler for cascades and the progress bars for it"""
    user_threads = args["Runtime"]["simultaneous_downloads_per_domain"]

    progress_table = await get_cascade_table(args["Progress_Options"])
    total_files = await Cascade.get_total()
    files = Files(overall_file_progress.add_task("[green]Completed", total=total_files),
                  overall_file_progress.add_task("[yellow]Skipped", total=total_files),
                  overall_file_progress.add_task("[red]Failed", total=total_files))
    with Live(progress_table, refresh_per_second=args["Progress_Options"]["refresh_rate"]):
        cascade_task = cascade_progress.add_task("[light_salmon3]Domains", progress_type="cascade",
                                                 total=len(Cascade.domains))

        downloaders = []
        tasks = []

        for domain, domain_obj in Cascade.domains.items():
            threads = user_threads if user_threads != 0 else multiprocessing.cpu_count()
            if 'bunkr' in domain or 'pixeldrain' in domain or 'anonfiles' in domain:
                threads = 2 if (threads > 2) else threads
            download_semaphore = asyncio.Semaphore(threads)
            downloaders.append(Downloader(args, client, SQL_Helper, scraper, threads, domain, domain_obj,
                                          download_semaphore, files))
        for downloader in downloaders:
            tasks.append(downloader.start_domain(cascade_task))
        await asyncio.gather(*tasks)

        cascade_progress.update(cascade_task, visible=False)

    await files.hide()

    await clear()
    await log(f"| [green]Files Complete: {files.completed_files}[/green] - [yellow]Files Skipped: "
              f"{files.skipped_files}[/yellow] - [red]Files Failed: {files.failed_files}[/red] |")


async def download_forums(args: dict, Forums: ForumItem, SQL_Helper: SQLHelper, client: Client,
                          scraper: ScrapeMapper) -> None:
    """Handler for forum threads and the progress bars for it"""
    user_threads = args["Runtime"]["simultaneous_downloads_per_domain"]

    progress_table = await get_forum_table(args["Progress_Options"])
    total_files = await Forums.get_total()
    files = Files(overall_file_progress.add_task("[green]Completed", total=total_files),
                  overall_file_progress.add_task("[yellow]Skipped", total=total_files),
                  overall_file_progress.add_task("[red]Failed", total=total_files))
    with Live(progress_table, refresh_per_second=args["Progress_Options"]["refresh_rate"]):
        forum_task = forum_progress.add_task("[orange3]FORUM THREADS", total=len(Forums.threads))
        for title, Cascade in Forums.threads.items():
            cascade_task = cascade_progress.add_task("[light_salmon3]" + title.upper(), total=len(Cascade.domains))

            downloaders = []
            tasks = []

            for domain, domain_obj in Cascade.domains.items():
                threads = user_threads if user_threads != 0 else multiprocessing.cpu_count()
                if 'bunkr' in domain or 'pixeldrain' in domain or 'anonfiles' in domain:
                    threads = 2 if (threads > 2) else threads
                download_semaphore = asyncio.Semaphore(threads)
                downloaders.append(Downloader(args, client, SQL_Helper, scraper, threads, domain, domain_obj,
                                              download_semaphore, files))
            for downloader in downloaders:
                tasks.append(downloader.start_domain(cascade_task))
            await asyncio.gather(*tasks)
            cascade_progress.update(cascade_task, visible=False)
            forum_progress.advance(forum_task, 1)

    await clear()
    await log(f"| [green]Files Complete: {files.completed_files}[/green] - [yellow]Files Skipped: "
              f"{files.skipped_files}[/yellow] - [red]Files Failed: {files.failed_files}[/red] |")
